"use strict";

var tap = require("tap");
var Pool = require("../..").Pool;
var utils = require("../utils");
var Promise = require("bluebird");
var ResourceFactory = utils.ResourceFactory;

tap.test("pool expands only to max limit", function(t) {
  var resourceFactory = new ResourceFactory();

  var factory = {
    name: "test1",
    create: resourceFactory.create.bind(resourceFactory),
    destroy: resourceFactory.destroy.bind(resourceFactory),
    validate: resourceFactory.validate.bind(resourceFactory),
    max: 1,
    min: 0,
    idleTimeoutMillis: 100,
    acquireTimeoutMillis: 100
  };

  var pool = new Pool(factory);

  pool
    .acquire()
    .then(obj => {
      return pool.acquire().catch(e => {
        t.ok(e instanceof Promise.TimeoutError);
        t.ok(e.message === "Operation timeout");
        pool.release(obj);
        t.end();
      });
    })
    .catch(t.threw);
});

tap.test("removes correct object on reap", function(t) {
  var resourceFactory = new ResourceFactory();

  var pool = new Pool({
    name: "test3",
    create: resourceFactory.create.bind(resourceFactory),
    destroy: resourceFactory.destroy.bind(resourceFactory),
    validate: resourceFactory.validate.bind(resourceFactory),
    max: 2,
    min: 0
  });

  pool.acquire().then(client => {
    // should be removed second
    setTimeout(function() {
      pool.destroy(client);
    }, 5);
  });

  pool.acquire().then(client => {
    // should be removed first
    pool.destroy(client);
  });

  setTimeout(function() {
    t.equal(1, resourceFactory.bin[0].id);
    t.equal(0, resourceFactory.bin[1].id);
    t.end();
  }, 100);
});

tap.test("drains", function(t) {
  var count = 5;
  var acquired = 0;

  var resourceFactory = new ResourceFactory();

  var pool = new Pool({
    name: "test4",
    create: resourceFactory.create.bind(resourceFactory),
    destroy: resourceFactory.destroy.bind(resourceFactory),
    validate: resourceFactory.validate.bind(resourceFactory),
    max: 2,
    min: 0,
    idleTimeoutMillis: 10000
  });

  // request 5 resources that release after 250ms
  for (var i = 0; i < count; i++) {
    pool.acquire().then(client => {
      acquired += 1;
      t.equal(typeof client.id, "number");
      setTimeout(() => {
        pool.destroy(client);
      }, 250);
    });
  }

  // ensure there are objects in pending queue
  // so drain can clear both pending and in-use queues
  t.notEqual(count, acquired);

  pool
    .drain()
    .then(() => {
      t.equal(count, acquired);
      t.equal(pool.available, 0);
      t.equal(pool.waiting, 0);
      t.equal(pool.using, 0);

      return pool.acquire().catch(e => {
        t.ok(e.message === "pool is draining and cannot accept work");
        t.end();
      });
    })
    .catch(t.threw);
});

tap.test("logging", function(t) {
  var logLevels = { verbose: 0, info: 1, warn: 2, error: 3 };
  var logMessages = { verbose: [], info: [], warn: [], error: [] };
  var resourceFactory = new ResourceFactory();

  var factory = {
    name: "test12",
    create: resourceFactory.create.bind(resourceFactory),
    destroy: () => {},
    validate: () => {},
    max: 2,
    min: 0,
    idleTimeoutMillis: 100,
    log: function(msg, level) {
      testlog(msg, level);
    }
  };
  var testlog = function(msg, level) {
    t.ok(level in logLevels);
    logMessages[level].push(msg);
  };

  var pool = new Pool(factory);
  var pool2 = new Pool({
    name: "testNoLog",
    create: resourceFactory.create.bind(resourceFactory),
    destroy: () => {},
    validate: () => {},
    max: 2,
    min: 0,
    idleTimeoutMillis: 100
  });

  t.equal(pool2.name, "testNoLog");

  pool
    .acquire()
    .then(() => {
      t.equal(
        logMessages.verbose[0],
        "createResource() - creating obj - count=1 min=0 max=2"
      );
      t.equal(logMessages.info[0], "dispense() clients=1 available=0");
      t.end();
    })
    .catch(t.threw);
});

tap.test("removes from available objects on destroy", function(t) {
  var destroyCalled = false;
  var factory = {
    name: "test13",
    create: function() {
      return Promise.resolve({});
    },
    destroy: function() {
      destroyCalled = true;
    },
    validate: () => {},
    max: 2,
    min: 0,
    idleTimeoutMillis: 100
  };

  var pool = new Pool(factory);

  pool
    .acquire()
    .then(obj => {
      pool.destroy(obj);
    })
    .then(() => {
      t.equal(destroyCalled, true);
      t.equal(pool.available, 0);
      t.equal(pool.using, 0);
      t.equal(pool.waiting, 0);
      t.end();
    });
});

tap.test("removes from available objects on validation failure", function(t) {
  var destroyCalled = 0;
  var validateCalled = 0;
  var destroyedClient = null;
  var count = 0;

  var factory = {
    name: "test14",
    create: () => {
      return Promise.resolve({ count: count++ });
    },
    destroy: client => {
      destroyCalled++;
      destroyedClient = client;
    },
    validate: client => {
      validateCalled++;
      return client.count > 0;
    },
    max: 2,
    min: 0,
    idleTimeoutMillis: 100
  };

  var pool = new Pool(factory);

  pool
    .acquire()
    .then(obj => {
      pool.release(obj);
      t.equal(obj.count, 0);

      return pool.acquire().then(obj => {
        pool.release(obj);
        t.equal(obj.count, 1);
      });
    })
    .then(() => {
      t.equal(validateCalled, 1);

      t.equal(destroyCalled, 1);
      t.equal(destroyedClient.count, 0); // destroy is called on first released client
      t.equal(pool.available, 1);

      t.end();
    })
    .catch(t.threw);
});

tap.test("acquire resolves after some failures", function(t) {
  var rejected = 0;

  var factory = {
    name: "test16",
    create: function() {
      rejected++;

      if (rejected <= 2) {
        return Promise.reject(new Error("Create error"));
      } else {
        return Promise.resolve({});
      }
    },
    destroy: () => {},
    validate: () => {},
    max: 1,
    min: 0
  };

  var pool = new Pool(factory);

  Promise.resolve()
    .then(() => t.rejects(pool.acquire(), new Error("Create error")))
    .then(() => t.rejects(pool.acquire(), new Error("Create error")))
    .then(() => pool.acquire())
    .then(() => {
      t.equal(pool.available, 0);
      t.end();
    })
    .catch(t.threw);
});

tap.test("returns only valid object to the pool", function(t) {
  var pool = new Pool({
    name: "test17",
    create: function() {
      return Promise.delay(1).then(() => ({ id: "validId" }));
    },
    destroy: () => {},
    validate: () => {},
    max: 1,
    min: 0,
    idleTimeoutMillis: 100
  });

  pool
    .acquire()
    .then(obj => {
      t.equal(pool.available, 0);
      t.equal(pool.using, 1);

      // Invalid release
      pool.release({});
      t.equal(pool.available, 0);
      t.equal(pool.using, 1);

      // Valid release
      pool.release(obj);
      t.equal(pool.available, 1);
      t.equal(pool.using, 0);

      t.end();
    })
    .catch(t.threw);
});
